@@ -1,4 +1,3 @@
-import six
 from astropy import units as u
 from astropy.io import fits
 from astropy import constants
@@ -11,7 +10,7 @@ from astropy.modeling.models import Ellipse2D, Gaussian2D
 from astropy.convolution import Kernel2D
 from astropy.convolution.kernels import _round_up_to_odd_integer
 
-from .utils import deconvolve, convolve, RadioBeamDeprecationWarning
+from .utils import deconvolve_optimized, convolve, RadioBeamDeprecationWarning
 
 # Conversion between a twod Gaussian FWHM**2 and effective area
 FWHM_TO_AREA = 2*np.pi/(8*np.log(2))
@@ -28,6 +27,16 @@ unit_format = {u.deg: r'\\circ',
                u.arcmin: "'"}
 
 
+def _with_default_unit(type_str, value, unit):
+    if not hasattr(value, 'unit'):
+        return value * unit
+
+    if value.unit.is_equivalent(unit):
+        return value
+    else:
+        raise u.UnitsError(f"{value.unit} for {type_str} is not equivalent to {equiv_unit}")
+
+
 class Beam(u.Quantity):
     """
     An object to handle single radio beams.
@@ -52,6 +61,8 @@ class Beam(u.Quantity):
             Gaussian beam.
         default_unit : :class:`~astropy.units.Unit`
             The unit to impose on major, minor if they are specified as floats
+        meta : dict, optional
+            A dictionary of metadata to store with the beam.
         """
 
         # improve to some kwargs magic later
@@ -63,36 +74,27 @@ class Beam(u.Quantity):
             if major is not None:
                 raise ValueError("Can only specify one of {major,minor,pa} "
                                  "and {area}")
-            rad = np.sqrt(area/(2*np.pi)) * u.deg
+            if not area.unit.is_equivalent(u.sr):
+                raise ValueError("Area unit should be equivalent to steradian.")
+            rad = np.sqrt(area/(2*np.pi))
             major = rad * SIGMA_TO_FWHM
             minor = rad * SIGMA_TO_FWHM
             pa = 0.0 * u.deg
 
         # give specified values priority
         if major is not None:
-            if u.deg.is_equivalent(major):
-                major = major
-            else:
-                warnings.warn("Assuming major axis has been specified in degrees")
-                major = major * u.deg
-        if minor is not None:
-            if u.deg.is_equivalent(minor):
-                minor = minor
-            else:
-                warnings.warn("Assuming minor axis has been specified in degrees")
-                minor = minor * u.deg
+            major = _with_default_unit("major", major, default_unit)
+
         if pa is not None:
-            if u.deg.is_equivalent(pa):
-                pa = pa
-            else:
-                warnings.warn("Assuming position angle has been specified in degrees")
-                pa = pa * u.deg
+            pa = _with_default_unit("pa", pa, u.deg)
         else:
-            pa = 0.0 * u.deg
+            pa = 0 * u.deg
 
         # some sensible defaults
         if minor is None:
             minor = major
+        else:
+            minor = _with_default_unit("minor", minor, default_unit)
 
         if minor > major:
             raise ValueError("Minor axis greater than major axis.")
@@ -113,7 +115,7 @@ class Beam(u.Quantity):
         return self
 
     @classmethod
-    def from_fits_bintable(cls, bintable, tolerance=0.01):
+    def from_fits_bintable(cls, bintable, tolerance=0.01, unit=u.arcsec):
         """
         Instantiate a single beam from a bintable from a CASA-produced image
         HDU.  The beams in the BinTableHDU will be averaged to form a single
@@ -150,11 +152,11 @@ class Beam(u.Quantity):
             warnings.warn("Metadata was averaged for keywords "
                           "{0}".format(",".join([key for key in meta])))
 
-        return cls(major=bmaj.mean()*u.arcsec, minor=bmin.mean()*u.arcsec,
+        return cls(major=bmaj.mean()*unit, minor=bmin.mean()*unit,
                    pa=circmean(bpa*u.deg, weights=bmaj/bmin))
 
     @classmethod
-    def from_fits_header(cls, hdr):
+    def from_fits_header(cls, hdr, unit=u.deg):
         """
         Instantiate the beam from a header. Attempts to extract the
         beam from standard keywords. Failing that, it looks for an
@@ -163,7 +165,7 @@ class Beam(u.Quantity):
         # ... given a file try to make a fits header
         # assume a string refers to a filename on disk
         if not isinstance(hdr,fits.Header):
-            if isinstance(hdr, six.string_types):
+            if isinstance(hdr, str):
                 if hdr.lower().endswith(('.fits', '.fits.gz', '.fit',
                                          '.fit.gz', '.fits.Z', '.fit.Z')):
                     hdr = fits.getheader(hdr)
@@ -176,7 +178,7 @@ class Beam(u.Quantity):
         # If we find a major axis keyword then we are in keyword
         # mode. Else look to see if there is an AIPS header.
         if "BMAJ" in hdr:
-            major = hdr["BMAJ"] * u.deg
+            major = hdr["BMAJ"] * unit
         else:
             hist_beam = cls.from_fits_history(hdr)
             if hist_beam is not None:
@@ -187,11 +189,11 @@ class Beam(u.Quantity):
         # Fill out the minor axis and position angle if they are
         # present. Else they will default .
         if "BMIN" in hdr:
-            minor = hdr["BMIN"] * u.deg
+            minor = hdr["BMIN"] * unit
         else:
             minor = None
         if "BPA" in hdr:
-            pa = hdr["BPA"] * u.deg
+            pa = hdr["BPA"] * unit
         else:
             pa = None
 
@@ -382,15 +384,22 @@ class Beam(u.Quantity):
             failure_returns_pointlike
         """
 
-        new_major, new_minor, new_pa = \
-            deconvolve(self, other,
-                       failure_returns_pointlike=failure_returns_pointlike)
+        new_major, new_minor, new_pa = deconvolve_optimized(self.to_header_keywords(),
+                                                            other.to_header_keywords(),
+                                                            failure_returns_pointlike=failure_returns_pointlike)
+
+        # Keep the units from before
+        new_major = (new_major * u.deg).to(self.major.unit)
+        new_minor = (new_minor * u.deg).to(self.minor.unit)
+        new_pa = (new_pa * u.rad).to(self.pa.unit)
+
         return Beam(major=new_major, minor=new_minor, pa=new_pa)
 
+
     def __eq__(self, other):
 
         # Catch floating point issues
-        atol_deg = 1e-12 * u.deg
+        atol_deg = 1e-10 * u.deg
 
         this_pa = self.pa.to(u.deg) % (180.0 * u.deg)
         other_pa = other.pa.to(u.deg) % (180.0 * u.deg)
@@ -499,7 +508,28 @@ class Beam(u.Quantity):
 
         return value.to(u.K, self.jtok_equiv(freq))
 
-    def ellipse_to_plot(self, xcen, ycen, pixscale):
+    @property
+    def beamarea_equiv(self):
+        return u.beam_angular_area(self.sr)
+
+    def commonbeam_with(self, other_beam):
+        '''
+        Solve for the common beam with a given `~radio_beam.Beam`.
+
+        This common beam operation is only valid for a set of 2 beams.
+        For the general case, define a set of beams using `~radio_beam.Beams`.
+
+        Parameters
+        ----------
+        other_beam : radio_beam.Beam
+            The beam to find the common beam with.
+
+        '''
+        from .commonbeam import find_commonbeam_between
+
+        return find_commonbeam_between(self, other_beam)
+
+    def ellipse_to_plot(self, xcen, ycen, pixscale, **kwargs):
         """
         Return a matplotlib ellipse for plotting
 
@@ -523,7 +553,7 @@ class Beam(u.Quantity):
                        height=(self.minor.to(u.deg) / pixscale).to(u.dimensionless_unscaled).value,
                        # PA is 90 deg offset from x-y axes by convention
                        # (it is angle from NCP)
-                       angle=(self.pa+90*u.deg).to(u.deg).value)
+                       angle=(self.pa+90*u.deg).to(u.deg).value, **kwargs)
 
     def as_kernel(self, pixscale, **kwargs):
         """
