@@ -10,10 +10,10 @@ except ImportError:
     HAS_SCIPY = False
 
 from .beam import Beam
-from .utils import BeamError, transform_ellipse
+from .utils import BeamError, transform_ellipse, deconvolve_optimized
 
 __all__ = ['commonbeam', 'common_2beams', 'getMinVolEllipse',
-           'common_manybeams_mve']
+           'common_manybeams_mve', 'find_commonbeam_between']
 
 
 def commonbeam(beams, method='pts', **method_kwargs):
@@ -42,8 +42,7 @@ def commonbeam(beams, method='pts', **method_kwargs):
         else:
             raise ValueError("method must be 'pts' or 'opt'.")
 
-
-def common_2beams(beams):
+def common_2beams(beams, check_deconvolution=True):
     '''
     Find a common beam from a `Beams` object with 2 beams. This
     function is based on the CASA implementation `ia.commonbeam`. Note that
@@ -60,32 +59,62 @@ def common_2beams(beams):
         The smallest common beam in the set of beams.
     '''
 
-    # This code is based on the implementation in CASA:
-    # https://open-bitbucket.nrao.edu/projects/CASA/repos/casa/browse/code/imageanalysis/ImageAnalysis/CasaImageBeamSet.cc
-
     if beams.size != 2:
         raise BeamError("This method is only valid for two beams.")
 
     if (~beams.isfinite).all():
         raise BeamError("All beams in the object are invalid.")
 
-    large_beam = beams.largest_beam()
+    return find_commonbeam_between(beams[0], beams[1], check_deconvolution=check_deconvolution)
+
+
+def find_commonbeam_between(beam1, beam2, check_deconvolution=True):
+    '''
+    Find the common beam between 2 `~radio_beam.Beam` objects.
+
+    This function is based on the CASA implementation `ia.commonbeam` that
+    the solution is valid when comparing 2 beams.
+
+    Parameters
+    ----------
+    beam1 : `~radio_beam.Beam`
+        Beam object.
+    beam2 : `~radio_beam.Beam`
+        Beam object.
+    check_deconvolution : bool, optional
+        Check that the common beam solution can be deconvolved from
+        both input beams.
+
+    Returns
+    -------
+    common_beam : `~radio_beam.Beam`
+        The smallest common beam in the set of beams.
+    '''
+
+    # This code is based on the implementation in CASA:
+    # https://open-bitbucket.nrao.edu/projects/CASA/repos/casa/browse/code/imageanalysis/ImageAnalysis/CasaImageBeamSet.cc
+
+    if not beam1.isfinite or not beam2.isfinite:
+        raise BeamError("At least one beam is invalid.")
+
+    # If equal, the common beam is itself.
+    if beam1 == beam2:
+        return beam1
+
+    if beam1 > beam2:
+        large_beam = beam1
+        small_beam = beam2
+    else:
+        large_beam = beam2
+        small_beam = beam1
+
     large_major = large_beam.major.to(u.arcsec)
     large_minor = large_beam.minor.to(u.arcsec)
 
-    if beams.argmax() == 0:
-        small_beam = beams[1]
-    else:
-        small_beam = beams[0]
     small_major = small_beam.major.to(u.arcsec)
     small_minor = small_beam.minor.to(u.arcsec)
 
-    # Case where they're already equal
-    if small_beam == large_beam:
-        return large_beam
-
-    deconv_beam = \
-        large_beam.deconvolve(small_beam, failure_returns_pointlike=True)
+    deconv_beam = large_beam.deconvolve(small_beam, failure_returns_pointlike=True)
 
     # Larger beam can be deconvolved. It is already the smallest common beam
     if deconv_beam.isfinite:
@@ -146,16 +175,17 @@ def common_2beams(beams):
                           minor=trans_minor + epsilon,
                           pa=trans_pa)
 
-        # Ensure this beam can now be deconvolved
-        deconv_large_beam = \
-            trans_beam.deconvolve(large_beam,
-                                  failure_returns_pointlike=True)
-        deconv_prob_beam = \
-            trans_beam.deconvolve(small_beam,
-                                  failure_returns_pointlike=True)
-        if not deconv_large_beam.isfinite or not deconv_prob_beam.isfinite:
-            raise BeamError("Failed to find common beam that both beams can "
-                            "be deconvolved by.")
+        if check_deconvolution:
+            # Ensure this beam can now be deconvolved
+            deconv_large_beam = \
+                trans_beam.deconvolve(large_beam,
+                                    failure_returns_pointlike=True)
+            deconv_prob_beam = \
+                trans_beam.deconvolve(small_beam,
+                                    failure_returns_pointlike=True)
+            if not deconv_large_beam.isfinite or not deconv_prob_beam.isfinite:
+                raise BeamError("Failed to find common beam that both beams can "
+                                "be deconvolved by.")
 
         # Taken from CASA implementation, but by adding epsilon, this shouldn't
         # be needed
@@ -336,13 +366,40 @@ def fits_in_largest(beams, large_beam=None):
     if large_beam is None:
         large_beam = beams.largest_beam()
 
-    for beam in beams:
-        if large_beam == beam:
-            continue
-        deconv_beam = large_beam.deconvolve(beam,
-                                            failure_returns_pointlike=True)
+    large_hdr_keywords = large_beam.to_header_keywords()
 
-        if not deconv_beam.isfinite:
+    majors = beams.major.to(u.deg).value
+    minors = beams.minor.to(u.deg).value
+    pas = beams.pa.to(u.deg).value
+
+    # Catch differences below  << 1 microarsec = 2.8e10
+    # This is the same limit used for checking equal beams in Beam.__eq__
+    atol_limit = 1e-12
+
+    for major, minor, pa in zip(majors, minors, pas):
+
+        equal = abs(large_hdr_keywords['BMAJ'] - major) < atol_limit
+        equal = equal and (abs(large_hdr_keywords['BMIN'] - minor) < atol_limit)
+
+        # Check if the beam is circular
+        # This checks for fractional changes below 1e-6 between the major and minor.
+        # Same limit used in Beam.__eq__
+        iscircular = (major - minor) / minor < 1e-6
+
+        # position angle only matters if the beam is asymmetric
+        if not iscircular:
+            equal = equal and (abs(((large_hdr_keywords['BPA'] % np.pi) - (pa % np.pi))) < atol_limit)
+
+        if equal:
+            continue
+
+        out = deconvolve_optimized(large_hdr_keywords,
+                                   {'BMAJ': major,
+                                    'BMIN': minor,
+                                    'BPA': pa},
+                                   failure_returns_pointlike=True)
+
+        if np.any([ax == 0. for ax in out[:2]]):
             return False
 
     return True
@@ -477,7 +534,11 @@ def ellipse_edges(beam, npts=300, epsilon=1e-3):
     return pts
 
 
-def common_manybeams_mve(beams, tolerance=1e-4, nsamps=200, epsilon=5e-4):
+def common_manybeams_mve(beams, tolerance=1e-4, nsamps=200,
+                         epsilon=5e-4,
+                         auto_increase_epsilon=True,
+                         max_epsilon=1e-3,
+                         max_iter=10):
     '''
     Calculate a common beam size using the Khachiyan Algorithm to find the
     minimum enclosing ellipse from all beam edges.
@@ -497,6 +558,17 @@ def common_manybeams_mve(beams, tolerance=1e-4, nsamps=200, epsilon=5e-4):
         the common beam can marginally be deconvolved for all beams. Small
         deviations result from the finite sampling of points and the choice
         of the tolerance.
+    auto_increase_epsilon : bool, optional
+        Re-run the algorithm when the solution cannot quite be deconvolved from
+        from all the beams. When `True`, epsilon is slightly increased with
+        each iteration until the common beam can be deconvolved from all beams.
+        Default is `True`.
+    max_epsilon : float, optional
+        Maximum epsilon value that is acceptable. Reached with `max_iter`.
+        Default is `1e-3`.
+    max_iter : int, optional
+        Maximum number of times to increase epsilon to try finding a valid
+        common beam solution.
 
     Returns
     -------
@@ -507,30 +579,52 @@ def common_manybeams_mve(beams, tolerance=1e-4, nsamps=200, epsilon=5e-4):
     if not HAS_SCIPY:
         raise ImportError("common_manybeams_mve requires scipy.optimize.")
 
-    pts = []
+    step = 1
 
-    for beam in beams:
-        pts.append(ellipse_edges(beam, nsamps, epsilon=epsilon))
+    while True:
+        pts = []
 
-    all_pts = np.hstack(pts).T
+        for beam in beams:
+            pts.append(ellipse_edges(beam, nsamps, epsilon=epsilon))
 
-    # Now find the outer edges of the convex hull.
-    hull = ConvexHull(all_pts)
-    edge_pts = all_pts[hull.vertices]
+        all_pts = np.hstack(pts).T
 
-    center, radii, rotation = \
-        getMinVolEllipse(edge_pts, tolerance=tolerance)
+        # Now find the outer edges of the convex hull.
+        hull = ConvexHull(all_pts)
+        edge_pts = all_pts[hull.vertices]
 
-    # The rotation matrix is coming out as:
-    # ((sin theta, cos theta)
-    #  (cos theta, - sin theta))
-    pa = np.arctan2(- rotation[0, 0], rotation[1, 0]) * u.rad
+        center, radii, rotation = \
+            getMinVolEllipse(edge_pts, tolerance=tolerance)
 
-    if pa.value == -np.pi or pa.value == np.pi:
-        pa = 0.0 * u.rad
+        # The rotation matrix is coming out as:
+        # ((sin theta, cos theta)
+        #  (cos theta, - sin theta))
+        pa = np.arctan2(- rotation[0, 0], rotation[1, 0]) * u.rad
 
-    com_beam = Beam(major=radii.max() * u.deg, minor=radii.min() * u.deg,
-                    pa=pa)
+        if pa.value == -np.pi or pa.value == np.pi:
+            pa = 0.0 * u.rad
+
+        com_beam = Beam(major=radii.max() * u.deg, minor=radii.min() * u.deg,
+                        pa=pa)
+
+        # If common beam is just slightly smaller than one of the beams,
+        # we increase epsilon to encourage a solution marginally larger
+        # so all beams can be convolved.
+        if auto_increase_epsilon:
+            if not fits_in_largest(beams, com_beam):
+                # Increase epsilon and run again
+                epsilon += (step + 1) * (max_epsilon - epsilon) / max_iter
+                step += 1
+
+                if step == max_iter + 1:
+                    raise BeamError("Could not increase epsilon to find"
+                                    " common beam.")
+
+                continue
+            else:
+                break
+        else:
+            break
 
     if not fits_in_largest(beams, com_beam):
         raise BeamError("Could not find common beam to deconvolve all beams.")
